@@ -10,6 +10,7 @@ import pandas as pd
 import plotly.io as pio
 import itertools
 
+from bravado_core.param import Param
 from bson.objectid import ObjectId
 from typing import Union, Type, List
 from tqdm.auto import tqdm
@@ -57,8 +58,11 @@ DEFAULT_HOST = "contribs-api.materialsproject.org"
 BULMA = "is-narrow is-fullwidth has-background-light"
 PROVIDERS = {"github", "google", "facebook", "microsoft", "amazon"}
 COMPONENTS = ["structures", "tables", "attachments"]  # using list to maintain order
-SUBDOMAINS = ["contribs", "lightsources", "ml"]
-LOCAL_STACK = {"contribs-api": 5000, "contribs-ml-api": 5002, "contribs-lightsources-api": 5003}
+SUBDOMAINS = ["contribs", "lightsources", "ml", "workshop-contribs"]
+LOCAL_STACK = {
+    "contribs-api": 5000, "workshop-contribs-api": 5005,
+    "contribs-ml-api": 5002, "contribs-lightsources-api": 5003,
+}
 VALID_URLS = {f"http://localhost:{p}" for p in LOCAL_STACK.values()}  # AWS Fargate Task
 VALID_URLS |= {f"http://{h}:{p}" for h, p in LOCAL_STACK.items()}  # local docker-compose stack
 VALID_URLS |= {f"https://{n}-api.materialsproject.org" for n in SUBDOMAINS}
@@ -336,8 +340,12 @@ def _run_futures(futures, total: int = 0, timeout: int = -1):
                 cnt = response.count if total_set and hasattr(response, "count") else 1
                 pbar.update(cnt)
 
-                if hasattr(future, "track_id") and hasattr(response, "result"):
-                    responses[future.track_id] = response.result
+                if hasattr(future, "track_id"):
+                    tid = future.track_id
+                    if hasattr(response, "result"):
+                        responses[tid] = response.result
+                    elif hasattr(response, "count"):
+                        responses[tid] = {"count": response.count}
 
                 elapsed = time.perf_counter() - start
                 timed_out = timeout > 0 and elapsed > timeout
@@ -446,48 +454,36 @@ class Client(SwaggerClient):
             # skip in tests
             return
 
-        columns = {"text": [], "number": []}
+        columns = {"string": [], "number": []}
+
         for project in resp["data"]:
             for column in project["columns"]:
                 if column["path"].startswith("data."):
                     col = column["path"].replace(".", "__")
                     if column["unit"] == "NaN":
-                        columns["text"].append(col)
+                        columns["string"].append(col)
                     else:
                         col = f"{col}__value"
                         columns["number"].append(col)
 
-        operators = {"text": ["exact", "contains"], "number": ["gte", "lte"]}
-        for path, d in spec_dict["paths"].items():
-            for verb in ["get", "put", "post", "delete"]:
-                if verb in d:
-                    old_params = deepcopy(d[verb].pop("parameters"))
-                    new_params, param_names = [], set()
+        resource = self.swagger_spec.resources["contributions"]
 
-                    while old_params:
-                        param = old_params.pop()
-                        if param["name"].startswith("^data__"):
-                            op = param["name"].rsplit("__", 1)[1]
-                            for typ, ops in operators.items():
-                                if op in ops:
-                                    for column in columns[typ]:
-                                        new_param = deepcopy(param)
-                                        param_name = f"{column}__{op}"
-                                        if param_name not in param_names:
-                                            new_param["name"] = param_name
-                                            desc = f"filter {column} via ${op}"
-                                            new_param["description"] = desc
-                                            new_params.append(new_param)
-                                            param_names.add(param_name)
-                        else:
-                            new_params.append(param)
+        for operation_id, operation in resource.operations.items():
+            for pn in list(operation.params.keys()):
+                if pn.startswith("data_"):
+                    param = operation.params.pop(pn)
+                    op = param.name.rsplit('$__', 1)[-1]
+                    typ = param.param_spec.get("type")
+                    key = "number" if typ == "number" else "string"
 
-                    d[verb]["parameters"] = new_params
-
-        swagger_spec = Spec.from_dict(spec_dict, origin_url, http_client, config)
-        super().__init__(
-            swagger_spec, also_return_response=bravado_config.also_return_response
-        )
+                    for column in columns[key]:
+                        param_name = f"{column}__{op}"
+                        param_spec = deepcopy(param.param_spec)
+                        param_spec["name"] = param_name
+                        param_spec.pop("description")
+                        operation.params[param_name] = Param(
+                            self.swagger_spec, operation, param_spec
+                        )
 
     def __dir__(self):
         members = set(self.swagger_spec.resources.keys())
@@ -571,13 +567,22 @@ class Client(SwaggerClient):
         track_id,
         params: dict,
         rel_url: str = "contributions",
-        op: str = "get"
+        op: str = "get",
+        data: dict = None
     ):
         if self.session and self.session.executor._shutdown:
             raise ValueError("Session closed. Use `with` statement.")
 
-        future = getattr(self.session, op)(
-            f"{self.url}/{rel_url}/", headers=self.headers, params=params
+        rname = rel_url.split("/", 1)[0]
+        resource = self.swagger_spec.resources[rname]
+        method = getattr(resource, f"{op}_entries").http_method
+        kwargs = dict(headers=self.headers, params=params)
+
+        if method == "put" and data:
+            kwargs["data"] = json.dumps(data).encode("utf-8")
+
+        future = getattr(self.session, method)(
+            f"{self.url}/{rel_url}/", **kwargs
         )
         setattr(future, "track_id", track_id)
         return future
@@ -722,9 +727,12 @@ class Client(SwaggerClient):
         values for the affected data fields will be respected.
 
         The `columns` argument is a dictionary which maps the data field names to its
-        units. Use `None` to indicate that a field doesn't have a unit. The unit for a
-        dimensionless quantity is an empty string (""). Nested fields are indicated using
-        a dot (".") in the data field name. Example:
+        units. Use `None` to indicate that a field is not a quantity (plain string). The
+        unit for a dimensionless quantity is an empty string (""). Percent (`%`) and
+        permille (`%%`) are considered units. Nested fields are indicated using a dot
+        (".") in the data field name.
+
+        Example:
 
         >>> client.init_columns("sandbox", {"a": None, "b.c": "eV", "b.d": "mm", "e": ""})
 
@@ -1083,7 +1091,13 @@ class Client(SwaggerClient):
 
         return ret
 
-    def update_contributions(self, name: str, data: dict, query: dict = None) -> dict:
+    def update_contributions(
+        self,
+        name: str,
+        data: dict,
+        query: dict = None,
+        timeout: int = -1
+    ) -> dict:
         """Apply the same update to all contributions in a project (matching query)
 
         See `client.contributions.get_entries()` for keyword arguments used in query.
@@ -1092,51 +1106,126 @@ class Client(SwaggerClient):
             name (str): name of the project
             data (dict): update to apply on every matching contribution
             query (dict): optional query to select contributions
+            timeout (int): cancel remaining requests if timeout exceeded (in seconds)
         """
         if not data:
             return "Nothing to update."
 
+        tic = time.perf_counter()
         valid = self._is_valid_payload("Contribution", data)
         if not valid:
             return {"error": valid}
 
-        _, per_page = self._get_per_page_default_max(op="update")
         query = query or {}
         query["project"] = name
-        has_more = True
-        updated = 0
+        cids = list(self.get_all_ids(query).get(name, {}).get("ids", set()))
 
-        while has_more:
-            resp = self.contributions.update_entries(
-                contributions=data, per_page=per_page, **query
-            ).result()
-            has_more = resp["has_more"]
-            updated += resp["count"]
+        if not cids:
+            print(f"There aren't any contributions to update for {name}")
+            return
 
-        print(f"Updated {updated} contributions.")
+        # get current list of data columns to decide if swagger reload is needed
+        resp = self.projects.get_entry(pk=name, _fields=["columns"]).result()
+        old_paths = set(c["path"] for c in resp["columns"])
 
-    def publish(self, name: str, recursive: bool = False) -> dict:
+        total = len(cids)
+        cids_query = {"id__in": cids}
+        _, total_pages = self.get_totals(query=cids_query)
+        queries = self._split_query(cids_query, op="update", pages=total_pages)
+        futures = [
+            self._get_future(i, q, op="update", data=data)
+            for i, q in enumerate(queries)
+        ]
+        responses = _run_futures(futures, total=total, timeout=timeout)
+        updated = sum(resp["count"] for _, resp in responses.items())
+
+        if updated:
+            resp = self.projects.get_entry(pk=name, _fields=["columns"]).result()
+            new_paths = set(c["path"] for c in resp["columns"])
+
+            if new_paths != old_paths:
+                self._load()
+
+        toc = time.perf_counter()
+        return {"updated": updated, "total": total, "seconds_elapsed": toc - tic}
+
+    def make_public(
+        self,
+        name: str,
+        query: dict = None,
+        recursive: bool = False,
+        timeout: int = -1
+    ) -> dict:
         """Publish a project and optionally its contributions
 
         Args:
             name (str): name of the project
+            query (dict): optional query to select contributions
             recursive (bool): also publish according contributions?
         """
+        return self._set_is_public(
+            True, name, query=query, recursive=recursive, timeout=timeout
+        )
+
+    def make_private(
+        self,
+        name: str,
+        query: dict = None,
+        recursive: bool = False,
+        timeout: int = -1
+    ) -> dict:
+        """Make a project and optionally its contributions private
+
+        Args:
+            name (str): name of the project
+            query (dict): optional query to select contributions
+            recursive (bool): also make according contributions private?
+        """
+        return self._set_is_public(
+            False, name, query=query, recursive=recursive, timeout=timeout
+        )
+
+    def _set_is_public(
+        self,
+        is_public: bool,
+        name: str,
+        query: dict = None,
+        recursive: bool = False,
+        timeout: int = -1
+    ) -> dict:
+        """Set the `is_public` flag for a project and optionally its contributions
+
+        Args:
+            is_public (bool): target value for `is_public` flag
+            name (str): name of the project
+            query (dict): optional query to select contributions
+            recursive (bool): also set `is_public` for according contributions?
+            timeout (int): cancel remaining requests if timeout exceeded (in seconds)
+        """
         try:
             resp = self.projects.get_entry(pk=name, _fields=["is_public"]).result()
         except HTTPNotFound:
             return {"error": f"project `{name}` not found or access denied!"}
 
-        is_public = resp["is_public"]
+        if not recursive and resp["is_public"] == is_public:
+            return {"warning": f"`is_public` already set to {is_public} for `{name}`."}
 
-        if not recursive and is_public:
-            return {"warning": f"project `{name}` already public (recursive=False)."}
+        ret = {}
 
-        if not is_public:
-            self.projects.update_entry(pk=name, project={"is_public": True}).result()
+        if resp["is_public"] != is_public:
+            resp = self.projects.update_entry(
+                pk=name, project={"is_public": is_public}
+            ).result()
+            ret["published"] = resp["is_public"] == is_public
 
         if recursive:
-            self.update_contributions(name, {"is_public": True}, {"is_public": False})
+            query = query or {}
+            query["is_public"] = not is_public
+            ret["contributions"] = self.update_contributions(
+                name, {"is_public": is_public}, query=query, timeout=timeout
+            )
+
+        return ret
 
     def submit_contributions(
         self,
@@ -1418,7 +1507,7 @@ class Client(SwaggerClient):
 
             end = datetime.utcnow()
             updated_total, _ = self.get_totals(query=dict(
-                project__in=project_names, last_modified__gt=start, last_modified__lt=end
+                project__in=project_names, last_modified__after=start, last_modified__before=end
             ), timeout=timeout)
             toc = time.perf_counter()
             dt = (toc - tic) / 60
